"""Routines to manipulate nested and flat dictionaries (and mix of both).

Used by :mod:`.process_routines` and :mod:`.config_routines`.
"""
import os
from typing import Any, Dict, Tuple, Union

import yaml
from flatten_dict import flatten as _flatten
from flatten_dict import unflatten as _unflatten

from cliconfig.yaml_tags._yaml_tags import get_yaml_loader, insert_tags


def merge_flat(
    dict1: Dict[str, Any],
    dict2: Dict[str, Any],
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Flatten then merge dict2 into dict1. The result is flat.

    Work even if dict1 and dict2 have a mix of nested and flat
    dictionaries. For instance like this:

    .. code-block:: python

        dict1 = {'a.b': 1, 'a': {'c': 2}, 'a.d': {'e.f': 3}}

    Parameters
    ----------
    dict1 : Dict[str, Any]
        The first dict. It can be nested, flat or a mix of both.
    dict2 : Dict[str, Any]
        The second dict to merge into first dict.
    allow_new_keys : bool, optional
        If True, new keys (that are not in dict1) are allowed in dict2.
        By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and dict2 has new keys that are not in dict1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.
        See last example. You may consider calling :func:`clean_pre_flat` on the input
        dicts in that case.

    Returns
    -------
    flat_dict : Dict[str, Any]
        The flat dict (all keys are at the root and separated by dots).

    Examples
    --------
    ::

        >>> merge_dict({'a.b': 1, 'a': {'c': 2}},  {'c': 3}, allow_new_keys=True)
        {'a.b': 1, 'a.c': 2, 'c': 3}
        >>> merge_dict({'a.b': 1, 'a': {'c': 2}},  {'c': 3}, allow_new_keys=False)
        ValueError: New parameter found 'c' that is not in the original dict.
        >>> merge_dict({'a.b': 1, 'a': {'b': 1}},  {'c': 3}, allow_new_keys=True)
        ValueError: duplicated key 'a.b'.
        The above exception was the direct cause of the following exception:
        ValueError: You may consider calling 'clean_pre_flat' on dict 1 before merging.
    """
    # Flatten dicts
    flat_dict1, flat_dict2 = _flat_before_merge(dict1, dict2)

    if not allow_new_keys:
        # Check that there are no new keys in dict2
        for key in flat_dict2:
            if key not in flat_dict1.keys():
                raise ValueError(
                    f"New parameter found '{key}' in that is not in the original "
                    "dict."
                )
    # Merge flat dicts
    flat_dict = {**flat_dict1, **flat_dict2}
    return flat_dict


def _flat_before_merge(
    dict1: Dict[str, Any], dict2: Dict[str, Any]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Flatten two dicts to merge them later."""
    # Flatten dicts
    flat_dicts = []
    for i, _dict in enumerate([dict1, dict2]):
        # Check if already flat
        is_flat = all(not isinstance(val, dict) for val in _dict.values())
        if not is_flat:
            try:
                flat_dict = flatten(_dict)
            except ValueError as exc:
                raise ValueError(
                    f"Duplicated key found in dict {i + 1} when flattening. "
                    f"You may consider calling 'clean_pre_flat' before merging."
                ) from exc
        else:
            flat_dict = _dict
        flat_dicts.append(flat_dict)
    return flat_dicts[0], flat_dicts[1]


def merge_flat_paths(
    dict_or_path1: Union[str, Dict[str, Any]],
    dict_or_path2: Union[str, Dict[str, Any]],
    *,
    allow_new_keys: bool = True,
) -> Dict[str, Any]:
    """Flatten then merge two dict eventually loaded from yaml file paths.

    Similar to :func:`.merge_flat` but allow passing the paths of dicts
    as inputs. It merges the second dict into the first one.

    Parameters
    ----------
    dict_or_path1 : Union[str, Dict[str, Any]]
        The first dict or its path.
    dict_or_path2 : Union[str, Dict[str, Any]]
        The second dict or its path, to merge into first dict.
    allow_new_keys : bool, optional
        If True, new keys (that are not in dict1) are allowed in dict2.
        By default True.

    Raises
    ------
    ValueError
        If allow_new_keys is False and dict2 has new keys that are not in dict1.
    ValueError
        If there are conflicting keys when flatten one of the dicts.
        See last example. You may consider calling :func:`clean_pre_flat` on the input
        dicts in that case.

    Returns
    -------
    flat_dict : Dict[str, Any]
        The flat dict (all keys are at the root and separated by dots).
    """
    dicts = []
    for dict_or_path in [dict_or_path1, dict_or_path2]:
        if isinstance(dict_or_path, str):
            _dict = load_dict(dict_or_path)
        else:
            _dict = dict_or_path
        dicts.append(_dict)
    dict1, dict2 = dicts[0], dicts[1]
    flat_dict = merge_flat(
        dict1,
        dict2,
        allow_new_keys=allow_new_keys,
    )
    return flat_dict


def flatten(in_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Flatten dict then return it (flat keys are built with dots).

    Work even if in_dict is a mix of nested and flat dictionaries.
    For instance like this:

    .. code-block:: python

        >>> flatten({'a.b': {'c': 1}, 'a': {'b.d': 2}, 'a.e': {'f.g': 3}})
        {'a.b.c': 1, 'a.b.d': 2, 'a.e.f.g': 3}


    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to flatten. It can be nested, already flat or a mix of both.

    Raises
    ------
    ValueError
        If dict has some conflicting keys (like ``{'a.b': <x>, 'a': {'b': <y>}}``).

    Returns
    -------
    flat_dict : Dict[str, Any]
        The flattened dict.

    Note
    ----
        Nested empty dict are ignored even if they are conflicting (see last example).

    Examples
    --------
    ::

        >>> flatten({'a.b': 1, 'a': {'c': 2}, 'd': 3})
        {'a.b': 1, 'a.c': 2, 'd': 3}
        >>> flatten({'a.b': {'c': 1}, 'a': {'b.d': 2}, 'a.e': {'f.g': 3}})
        {'a.b.c': 1, 'a.b.d': 2, 'a.e.f.g': 3}
        >>> flatten({'a.b': 1, 'a': {'b': 1}})
        ValueError: duplicated key 'a.b'
        >>> flatten({'a.b': 1, 'a': {'c': {}}, 'a.c': 3})
        {'a.b': 1, 'a.c': 3}

    """
    flat_dict = _flatten(in_dict, reducer="dot")
    return flat_dict


def unflatten(flat_dict: Dict[str, Any]) -> Dict[str, Any]:
    """Unflatten a flat dict then return it.

    Parameters
    ----------
    flat_dict : Dict[str, Any]
        The dict to unflatten. Must be a fully flat dict (depth of 1 with keys
        separated by dots).

    Raises
    ------
    ValueError
        If flat_dict is not flat and then found conflicts.

    Returns
    -------
    unflat_dict : Dict[str, Any]
        The output nested dict.

    Examples
    --------
    ::

        >>> unflatten({'a.b': 1, 'a.c': 2, 'c': 3})
        {'a': {'b': 1, 'c': 2}, 'c': 3}
        >>> unflatten({'a.b': 1, 'a': {'c': 2}})
        ValueError: duplicated key 'a'
        The dict must be flatten before calling unflatten function.
    """
    try:
        unflat_dict = _unflatten(flat_dict, splitter="dot")
    except ValueError as exc:
        raise ValueError(
            "The dict must be flatten before calling unflatten function."
        ) from exc
    return unflat_dict


def clean_pre_flat(in_dict: Dict[str, Any], priority: str) -> Dict[str, Any]:
    """Remove keys in input dict that may cause conflicts when flattening.

    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to clean. It must be the union of a fully flat dict
        (no nested dict i values) and a fully nested dict (without dots in keys).
        See warning section below.
    priority: str
        One of 'flat' or 'unflat', 'error'.
        If 'flat', keys with dots at the root like ``{'a.b': ...}`` (flat keys) have
        priority over nested keys like ``{'a': {'b': ...}}`` when there are conflicts.
        If 'unflat', nested keys have priority over flat keys when there are conflicts.
        If 'error', raise an error when there are conflicts.

    Raises
    ------
    ValueError
        If priority is not one of 'flat', 'unflat' or 'error'.

    Returns
    -------
    Dict[str, Any]
        The cleansed dict.

    Warning
    -------
        * No flat key can contain a dict. Then, dicts like ``{'a.b': {'c': 1}}``
          are not supported.
        * All the keys that contain dots (the flat keys) must be at the root.
          Then, dicts like ``{a: {'b.c': 1}}`` are not supported.
        * To summarize, the dict must contain only fully flat dicts
          and/or fully nested dicts.

    Examples
    --------
    ::

        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='flat')
        {'a.b': 1, 'c': 3}
        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='unflat')
        {'a': {'b': 2}, 'c': 3}
        >>> clean_pre_flat({'a.b': 1, 'a': {'b': 2}, 'c': 3}, priority='error')
        ValueError: duplicated key 'a.b'
    """
    if priority in ("flat", "unflat"):
        # Check that there are no conflicts
        flat_part = dict(filter(lambda items: "." in items[0], in_dict.items()))
        unflat_part = dict(filter(lambda items: "." not in items[0], in_dict.items()))
        unflat_part_flat = flatten(unflat_part)
        if priority == "unflat":
            for key in flat_part:
                if key in unflat_part_flat:
                    _del_key(in_dict, key, keep_flat=False, keep_unflat=True)
        else:
            for key in unflat_part_flat:
                if key in flat_part:
                    _del_key(in_dict, key, keep_flat=True, keep_unflat=False)
    elif priority == "error":
        flatten(in_dict)  # Will raise an error if there are conflicts
    else:
        raise ValueError(
            "priority argument must be one of 'flat', 'unflat' or 'error' but "
            f"found '{priority}'."
        )
    return in_dict


def _del_key(
    in_dict: Dict[str, Any],
    flat_key: str,
    *,
    keep_flat: bool = False,
    keep_unflat: bool = False,
) -> None:
    """Remove in place a value in dict corresponding to a flat key (e.g. 'a.b.c').

    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to clean. It must be the union of a fully flat dict
        (no nested dict i values) and a fully nested dict (without dots in keys).
        See warning section below.
    flat_key: str
        The flat key to remove. E.g. 'a.b.c'.
    keep_flat: bool, optional
        If True, keep the flat key in the dict. By default False.
    keep_unflat: bool, optional
        If True, keep the unflat key in the dict. By default False.

    Raises
    ------
    ValueError
        If the key is not found in the dict.

    Warning
    -------
        * No flat key can contain a dict. Then, dicts like ``{'a.b': {'c': 1}}``
          are not supported.
        * All the keys that contain dots (the flat keys) must be at the root.
          Then, dicts like ``{a: {'b.c': 1}}`` are not supported.
        * To summarize, the dict must contain only fully flat dicts
          and fully nested dicts.

    Examples
    --------
    ::

        >>> in_dict = {'a': {'b': {'c': 1}, 'd': 2}, 'a.b.c': 4}
        >>> _del_key(in_dict, 'a.b.c'); in_dict
        {'a': {'d': 2}}
        >>> _del_key(in_dict, 'a.b.c', keep_flat=True); in_dict
        {'a': {'d': 2}, 'a.b.c': 4}
        >>> _del_key(in_dict, 'a.b.c', keep_unflat=True); in_dict
        {'a': {'b': {'c': 1}, 'd': 2}}
        >>> _del_key(in_dict, 'a.b.z')
        ValueError: Key 'a.b.z' not found in dict.
        >>> _del_key(in_dict, 'a.z.c')
        ValueError: Key 'a.z.c' not found in dict.
    """
    found_key = False
    if not keep_flat and flat_key in in_dict:
        # Remove flat_key if it exists at the root
        found_key = True
        del in_dict[flat_key]

    def recursive_del_key(
        in_dict: Dict[str, Any], key: str, *, found_key: bool
    ) -> bool:
        first_key, *other_keys = key.split(".", 1)
        if other_keys:
            if first_key in in_dict:
                # Delete key in sub-dict
                new_key = recursive_del_key(
                    in_dict[first_key],
                    ".".join(other_keys),
                    found_key=found_key,
                )
                found_key = found_key or new_key
                # Remove if empty
                if in_dict[first_key] == {}:
                    del in_dict[first_key]
            else:
                # No key found, return input found_key
                return found_key
        else:
            if first_key in in_dict:
                found_key = True
                del in_dict[first_key]
            else:
                # No key found, return input found_key
                return found_key
        return found_key

    if not keep_unflat:
        # Remove flat_key if it exists in a nested dict
        found_key = recursive_del_key(in_dict, flat_key, found_key=found_key)
    # Raise error if key not found
    if not found_key:
        raise ValueError(f"Key '{flat_key}' not found in dict.")


def save_dict(in_dict: Dict[str, Any], path: str) -> None:
    """Save a dict to a yaml file (with yaml.dump).

    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to save.
    path : str
        The path to the yaml file to save the dict.
    """
    dir_path = os.path.dirname(path)
    os.makedirs(dir_path, exist_ok=True)
    with open(path, "w", encoding="utf-8") as cfg_file:
        yaml.dump(in_dict, cfg_file, default_flow_style=False)


def load_dict(path: str) -> Dict[str, Any]:
    """Load dict from a yaml file path.

     Support multiple files in the same document and yaml tags.

    Parameters
    ----------
    path : str
        The path to the file to load the dict.

    Returns
    -------
    out_dict : Dict[str, Any]
        The nested (unflatten) loaded dict.

    Note
    ----

        * If multiple yaml files are in the same document, they are merged
        from the first to the last.
        * To use multiple yaml tags, separate them with "@". E.g. ``!tag1@tag2``.
        * You can combine any number of yaml and cliconfig tags together.
    """
    with open(path, "r", encoding="utf-8") as cfg_file:
        file_dicts = yaml.load_all(cfg_file, Loader=get_yaml_loader())
        out_dict: Dict[str, Any] = {}
        for file_dict in file_dicts:
            new_dict, _ = insert_tags(file_dict)
            out_dict = merge_flat(out_dict, new_dict, allow_new_keys=True)
    return unflatten(out_dict)


def show_dict(in_dict: Dict[str, Any], start_indent: int = 0) -> None:
    """Show the input dict in a pretty way.

    The config dict is automatically unflattened before printing.

    Parameters
    ----------
    in_dict : Dict[str, Any]
        The dict to show.
    start_indent : int, optional
        The number of starting tab indent (4 spaces), by default 0.
    """

    def pretty_print(in_dict: Dict[str, Any], indent: int) -> None:
        """Pretty print the dict recursively."""
        for key, value in in_dict.items():
            print(f"{'    ' * indent}{key}: ", end="")
            if isinstance(value, dict):
                print()
                pretty_print(value, indent + 1)
            elif isinstance(value, str):
                print(f"'{value}'")
            else:
                print(value)

    pretty_print(unflatten(in_dict), start_indent)



"""Test the tag routines."""

import pytest
import pytest_check as check

from cliconfig.tag_routines import clean_all_tags, clean_tag, dict_clean_tags, is_tag_in


@pytest.fixture()
def key1() -> str:
    """Return a key with many tags."""
    return "abc@tag.def@tag_2.ghi@tag"


@pytest.fixture()
def key2() -> str:
    """Return a key with many many tags."""
    return "abc@hashtag@tagg@tag@tag 2@tag@ag.def@tag _.jkl@tag.mno@tag"


def test_clean_tag(key1: str, key2: str) -> None:
    """Test clean_tag."""
    check.equal(
        clean_tag(key1, "tag"),
        "abc.def@tag_2.ghi",
    )
    check.equal(
        clean_tag(key1, "@tag"),
        "abc.def@tag_2.ghi",
    )
    check.equal(clean_tag(key2, "tag"), "abc@hashtag@tagg@tag 2@ag.def@tag _.jkl.mno")
    check.equal(clean_tag(key2, "@tag"), "abc@hashtag@tagg@tag 2@ag.def@tag _.jkl.mno")


def test_clean_all_tags(key1: str, key2: str) -> None:
    """Test clean_all_tags."""
    check.equal(
        clean_all_tags(key1),
        "abc.def.ghi",
    )
    check.equal(
        clean_all_tags(key2),
        "abc.def.jkl.mno",
    )


def test_dict_clean_tags(key1: str, key2: str) -> None:
    """Test dict_clean_tag."""
    in_dict = {
        "abc.def.pqr": 0,
        key1: 1,
        key2: 2,
    }
    out_dict, tagged_keys = dict_clean_tags(in_dict)
    check.equal(out_dict, {"abc.def.pqr": 0, "abc.def.ghi": 1, "abc.def.jkl.mno": 2})
    check.is_true(key1 in tagged_keys)
    check.is_true(key1 in tagged_keys)


def test_is_tag_in() -> None:
    """Test is_tag_in."""
    check.is_true(is_tag_in("config.config2.config3@tag", "tag"))
    check.is_false(is_tag_in("config.config2.config3@tag", "tog"))
    check.is_false(is_tag_in("config.config2.config3@tag_2", "tag"))
    check.is_false(is_tag_in("config.config2.config3@tag_2@tog", "tag"))
    check.is_true(is_tag_in("config.config2.config3@tag@tog", "@tag"))
    check.is_true(is_tag_in("config.config2.config3@tag@tog", "@tog"))
    check.is_true(is_tag_in("config.config2@tag.config3@tog", "tag", full_key=True))
    check.is_false(is_tag_in("config.config2@tag.config3@tog", "tag", full_key=False))



"""Shared pytest fixtures."""
from typing import Any, Dict

import pytest

from cliconfig.base import Config
from cliconfig.processing.base import Processing
from cliconfig.tag_routines import clean_all_tags, clean_tag, is_tag_in


class ProcessAdd1(Processing):
    """Add 1 to values with tag "@add1"."""

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if is_tag_in(flat_key, "add1"):
                flat_config.dict[clean_tag(flat_key, "add1")] = value + 1
                del flat_config.dict[flat_key]
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing."""
        flat_config.dict["processing name"] = "ProcessAdd1"
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing."""
        return self.premerge(flat_config)


class ProcessKeep(Processing):
    """Prevent a value from being changed after the merge."""

    def __init__(self) -> None:
        super().__init__()
        self.keep_vals: Dict[str, Any] = {}

    # pylint: disable=unused-argument
    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if is_tag_in(flat_key, "keep"):
                new_key = clean_tag(flat_key, "keep")
                flat_config.dict[new_key] = value
                del flat_config.dict[flat_key]
                self.keep_vals[clean_all_tags(flat_key)] = value
        return flat_config

    # pylint: disable=unused-argument
    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        for key, value in self.keep_vals.items():
            if key in flat_config.dict:
                flat_config.dict[key] = value
        self.keep_vals = {}
        return flat_config

    # pylint: disable=unused-argument
    def postload(self, flat_config: Config) -> Config:
        """Post-load processing."""
        for key, value in self.keep_vals.items():
            flat_config.dict[key] = value
        self.keep_vals = {}
        return flat_config


@pytest.fixture()
def process_add1() -> ProcessAdd1:
    """Return a processing object that adds 1 on tag "@add1"."""
    return ProcessAdd1()


@pytest.fixture()
def process_keep() -> ProcessKeep:
    """Return a processing object that keep a value unchanged after the merge."""
    return ProcessKeep()



"""Base class for processing.

Used to make configuration object and run the routines in :mod:`.process_routines`
and :mod:`.config_routines`.
"""
from cliconfig.base import Config


class Processing:
    """Processing base class.

    Each processing classes contains pre-merge, post-merge, pre-save
    and post-load processing. They are used with routines that apply
    processing in :mod:`.process_routines` and
    :mod:`.config_routines`.

    That are applied in the order defined
    by the order attribute in case of multiple processing.
    """

    def __init__(self) -> None:
        self.premerge_order = 0.0
        self.postmerge_order = 0.0
        self.endbuild_order = 0.0
        self.presave_order = 0.0
        self.postload_order = 0.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing.

        Function applied to the flat config to modify it
        before merging. It takes a flat config and returns a flat config.
        """
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing.

        Function applied to the flat config to modify it
        after merging . It takes a flat config and returns a flat config.
        """
        return flat_config

    def endbuild(self, flat_config: Config) -> Config:
        """End-build processing.

        Function applied to the flat config to modify it at the end of
        a building process (typically :func:`.make_config` or :func:`.load_config`).
        It takes a flat config and returns a flat config.
        """
        return flat_config

    def presave(self, flat_config: Config) -> Config:
        """Pre-save processing.

        Function applied to the flat config to modify it before
        saving. It takes a flat config and returns a flat config.
        """
        return flat_config

    def postload(self, flat_config: Config) -> Config:
        """Post-load processing.

        Function applied to the flat config to modify it after
        loading. It takes a flat config and returns a flat config.
        """
        return flat_config

    def __eq__(self, __value: object) -> bool:
        """Equality operator.

        Two processing are equal if they are the same class and add the same
        attributes (accessed with ``__dict__``).
        """
        equal = (
            isinstance(__value, self.__class__) and self.__dict__ == __value.__dict__
        )
        return equal



"""Test ProcessBypassTyping."""

from typing import Dict, Set

import pytest
import pytest_check as check

from cliconfig import Config
from cliconfig.process_routines import merge_flat_processing
from cliconfig.processing.base import Processing
from cliconfig.processing.builtin import ProcessTyping
from cliconfig.tag_routines import clean_all_tags, clean_tag, is_tag_in


class ProcessPrintSorted(Processing):
    """Print the parameters tagged with "@look", sorted by value on post-merge."""

    def __init__(self) -> None:
        super().__init__()
        self.looked_keys: Set[str] = set()
        # Pre-merge just look for the tag so order is not important
        self.premerge_order = 0.0
        # Post-merge should be after the copy processing if we want the final values
        # on post-merge
        self.postmerge_order = 15.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        # Browse a freeze version of the dict (because we will modify it to remove tags)
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if is_tag_in(flat_key, "look"):  # Check if the key contains the tag
                # Remove the tag and update the dict
                new_key = clean_tag(flat_key, "look")
                flat_config.dict[new_key] = value
                del flat_config.dict[flat_key]
                # Store the key
                # remove all tags = true parameter name
                clean_key = clean_all_tags(flat_key)
                self.looked_keys.add(clean_key)
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        values = []
        for key in self.looked_keys:
            # IMPORTANT
            # ("if key in flat_config.dict:" is important in case of some keys were
            # removed or if multiple dicts with different parameters are seen by
            # the processing)
            if key in flat_config.dict:
                values.append(flat_config.dict[key])
        print("The sorted looked values are: ", sorted(values))
        # If we don't want to keep the looked keys for further print:
        self.looked_keys = set()

        return flat_config


class ProcessBypassTyping(Processing):
    """Bypass type check of ProcessTyping for parameters tagged with "@bypass_typing".

    In pre-merge it looks for a parameter with the tag "@bypass_typing",
    removes it and change the internal ProcessTyping variables to avoid
    checking the type of the parameter with ProcessTyping.
    """

    def __init__(self) -> None:
        super().__init__()
        self.bypassed_forced_types: Dict[str, tuple] = {}
        # Before ProcessTyping pre-merge to let it change the type
        self.premerge_order = -1.0

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if is_tag_in(flat_key, "bypass_typing"):
                new_key = clean_tag(flat_key, "bypass_typing")
                flat_config.dict[new_key] = value
                del flat_config.dict[flat_key]
                clean_key = clean_all_tags(flat_key)
                for processing in flat_config.process_list:
                    if (
                        isinstance(processing, ProcessTyping)
                        and clean_key in processing.forced_types
                    ):
                        forced_type = processing.forced_types.pop(clean_key)
                        self.bypassed_forced_types[clean_key] = forced_type
        return flat_config


def test_process_print_sorted(capsys: pytest.CaptureFixture) -> None:
    """Test ProcessPrintSorted."""
    config1 = Config({"a@look": 0, "b@look": 2, "c": 3}, [ProcessPrintSorted()])
    config2 = Config({"a@look": 1, "d@look": 4}, [])
    capsys.readouterr()
    merge_flat_processing(config1, config2)
    out = capsys.readouterr().out
    check.equal(out, "The sorted looked values are:  [1, 2, 4]\n")


def test_process_bypass_typing() -> None:
    """Test ProcessBypassTyping."""
    config1 = Config({"a@type:int": 0}, [ProcessBypassTyping(), ProcessTyping()])
    config2 = Config({"a@bypass_typing@type:str": "a"}, [])
    merge_flat_processing(config1, config2)
    config1 = Config({"a@type:int": 0}, [ProcessBypassTyping(), ProcessTyping()])
    config2 = Config({"a@type:str": "a"}, [])
    # Reset ProcessTyping
    with pytest.raises(
        ValueError,
        match=(
            "Find the tag '@type:str' on a key that has already "
            "been associated to an other type: int.*"
        ),
    ):
        merge_flat_processing(config1, config2)



"""Tests for config.py."""
import os
import shutil
import sys

import pytest
import pytest_check as check

from cliconfig.base import Config
from cliconfig.config_routines import load_config, make_config, save_config, show_config
from cliconfig.processing.base import Processing


def test_make_config(capsys: pytest.CaptureFixture, process_add1: Processing) -> None:
    """Test make_config."""
    sys_argv = sys.argv.copy()
    sys.argv = [
        "tests/test_make_config.py.py",
        "--config",
        "tests/configs/config1.yaml,tests/configs/config2.yaml",
        "--unknown",
        "--param2@add1=6",
        "--unknown2=8",  # check that not error but a warning in console
    ]
    capsys.readouterr()  # Clear stdout and stderr
    config = make_config(
        "tests/configs/default1.yaml",
        "tests/configs/default2.yaml",
        process_list=[process_add1],
    )
    captured = capsys.readouterr()
    out = captured.out
    expected_config = {
        "param1": 4,
        "param2": 7,
        "param3": 3,
        "letters": {
            "letter1": "f",
            "letter2": "e",
            "letter3": "c",
            "letter4": "d",
        },
        "processing name": "ProcessAdd1",
    }
    expected_out = (
        "[CONFIG] Warning: New keys found in CLI parameters that will not be merged:\n"
        "  - unknown\n"
        "  - unknown2\n"
        "[CONFIG] Info: Merged 2 default config(s), "
        "2 additional config(s) and 1 CLI parameter(s).\n"
    )
    check.equal(config.dict, expected_config)
    check.equal(out, expected_out)

    # No additional configs
    sys.argv = [
        "tests/test_make_config.py.py",
    ]
    config = make_config("tests/configs/default1.yaml", "tests/configs/default2.yaml")
    expected_config = {
        "param1": 1,
        "param2": 2,
        "param3": 3,
        "letters": {
            "letter1": "a",
            "letter2": "b",
            "letter3": "c",
            "letter4": "d",
        },
    }
    check.equal(config.dict, expected_config)
    config = make_config(add_default_processing=False)
    check.equal(config.dict, {})
    check.equal(config.process_list, [])

    sys.argv = sys_argv.copy()


def test_load_config(process_add1: Processing) -> None:
    """Test and load_config."""
    # With default configs
    config = load_config(
        "tests/configs/config2.yaml",
        default_config_paths=[
            "tests/configs/default1.yaml",
            "tests/configs/default2.yaml",
        ],
        process_list=[process_add1],
    )
    expected_config = {
        "param1": 4,
        "param2": 2,
        "param3": 3,
        "letters": {
            "letter1": "a",
            "letter2": "e",
            "letter3": "c",
            "letter4": "d",
        },
        "processing name": "ProcessAdd1",
    }
    check.equal(config.dict, expected_config)
    # Additional keys when allow_new_keys=False
    with pytest.raises(ValueError, match="New parameter found 'param3'.*"):
        load_config(
            "tests/configs/default2.yaml",
            default_config_paths=[
                "tests/configs/default1.yaml",
            ],
        )


def test_show_config() -> None:
    """Test show_config."""
    show_config(Config({"a": 1, "b": {"c": 2, "d": 3}, "e": "f"}, []))


def test_save_config() -> None:
    """Test save_config."""
    config = Config({"a": 1, "b": {"c": 2, "d": 3}, "e": "f"}, [])
    save_config(config, "tests/tmp/config.yaml")
    check.is_true(os.path.exists("tests/tmp/config.yaml"))
    shutil.rmtree("tests/tmp")



"""Base classes of Config object."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional

# Imports Processing for mypy type checking only
if TYPE_CHECKING:
    from cliconfig.processing.base import Processing


class Config:
    """Class for configuration.

    Config object contain the config dict and the processing list
    and no methods except ``__init__``, ``__repr__`` and ``__eq__``.
    The Config objects are mutable and not hashable.

    Parameters
    ----------
    config_dict : Dict[str, Any]
        The config dict.
    process_list : Optional[List[Processing]], optional
        The list of Processing objects. If None, an empty list is used.
        The default is None.
    """

    def __init__(
        self,
        config_dict: Dict[str, Any],
        process_list: Optional[List["Processing"]] = None
    ) -> None:
        self.dict = config_dict
        self.process_list = process_list if process_list else []

    def __repr__(self) -> str:
        """Representation of Config object."""
        process_classes = [process.__class__.__name__ for process in self.process_list]
        return f"Config({self.dict}, {process_classes})"

    def __eq__(self, other: Any) -> bool:
        """Equality operator.

        Two Config objects are equal if their dicts are equal and their
        lists of Processing objects are equal (order doesn't matter).
        """
        if (
            isinstance(other, Config)
            and self.dict == other.dict
            and len(self.process_list) == len(other.process_list)
        ):
            equal = True
            for processing in self.process_list:
                equal = equal and processing in other.process_list
            for processing in other.process_list:
                equal = equal and processing in self.process_list
            return equal
        return False



"""Tests for dict routines."""
import os
import shutil

import pytest
import pytest_check as check

from cliconfig.dict_routines import (
    _del_key,
    clean_pre_flat,
    flatten,
    load_dict,
    merge_flat,
    merge_flat_paths,
    save_dict,
    show_dict,
    unflatten,
)


def test_flatten() -> None:
    """Test flatten(."""
    check.equal(
        flatten({"a.b": 1, "a": {"c": 2}, "d": 3}),
        {"a.b": 1, "a.c": 2, "d": 3},
    )
    check.equal(
        flatten({"a.b": {"c": 1}, "a": {"b.d": 2}, "a.e": {"f.g": 3}}),
        {"a.b.c": 1, "a.b.d": 2, "a.e.f.g": 3},
    )
    check.equal(
        flatten({"a.b": 1, "a": {"c": {}}, "a.c": 3}),
        {"a.b": 1, "a.c": 3},
    )
    with pytest.raises(ValueError, match="duplicated key 'a.b'"):
        flatten({"a.b": 1, "a": {"b": 1}})


def test_unflatten() -> None:
    """Test unflatten."""
    check.equal(
        unflatten({"a.b": 1, "a.c": 2, "c": 3}),
        {"a": {"b": 1, "c": 2}, "c": 3},
    )
    with pytest.raises(ValueError, match="The dict must be flatten.*"):
        unflatten({"a.b": 1, "a": {"c": 2}})


def test_merge_flat() -> None:
    """Test merge_flat."""
    dict1 = {"a.b": 1, "a": {"c": 2}}
    dict2 = {"c": 3}
    check.equal(
        merge_flat(dict1, dict2, allow_new_keys=True),
        {"a.b": 1, "a.c": 2, "c": 3},
    )
    with pytest.raises(ValueError, match="New parameter found 'c'.*"):
        merge_flat(dict1, dict2, allow_new_keys=False)
    dict1 = {"a.b": 1, "a": {"b": 1, "c": 2}}
    with pytest.raises(
        ValueError,
        match="Duplicated key found.*You may consider calling 'clean_pre_flat'.*",
    ):
        merge_flat(dict1, dict2, allow_new_keys=True)


def test_merge_flat_paths() -> None:
    """Test merge_flat_paths."""
    dict1 = {
        "param1": 1,
        "param2": 2,
        "letters": {
            "letter1": "a",
            "letter2": "b",
        },
    }
    dict2 = {
        "param3": 3,
        "letters": {
            "letter3": "c",
            "letter4": "d",
        },
    }
    expected_dict = {
        "param1": 1,
        "param2": 2,
        "param3": 3,
        "letters.letter1": "a",
        "letters.letter2": "b",
        "letters.letter3": "c",
        "letters.letter4": "d",
    }

    flat_dict = merge_flat_paths(
        "tests/configs/default1.yaml",
        "tests/configs/default2.yaml",
        allow_new_keys=True,
    )
    check.equal(flat_dict, expected_dict)
    flat_dict = merge_flat_paths(dict1, dict2, allow_new_keys=True)
    check.equal(flat_dict, expected_dict)
    flat_dict = merge_flat_paths(
        dict1, "tests/configs/default2.yaml", allow_new_keys=True
    )
    check.equal(flat_dict, expected_dict)
    with pytest.raises(ValueError, match="New parameter found 'param3'.*"):
        merge_flat_paths(
            "tests/configs/default1.yaml",
            "tests/configs/default2.yaml",
            allow_new_keys=False,
        )


def test_del_key() -> None:
    """Test _del_key."""
    in_dict = {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}
    _del_key(in_dict, "a.b.c")
    check.equal(in_dict, {"a": {"d": 2}, "a.e": 3})

    in_dict = {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}
    _del_key(in_dict, "a.b.c", keep_flat=True)
    check.equal(in_dict, {"a": {"d": 2}, "a.e": 3, "a.b.c": 4})

    in_dict = {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}
    _del_key(in_dict, "a.b.c", keep_unflat=True)
    check.equal(in_dict, {"a": {"b": {"c": 1}, "d": 2}, "a.e": 3})

    with pytest.raises(ValueError, match="Key 'a.b.z' not found in dict."):
        _del_key({"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}, "a.b.z")
    with pytest.raises(ValueError, match="Key 'a.z.c' not found in dict."):
        _del_key({"a": {"b": {"c": 1}, "d": 2}, "a.e": 3, "a.b.c": 4}, "a.z.c")


def test_clean_pre_flat() -> None:
    """Test clean_pre_flat."""
    check.equal(
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="flat"),
        {"a.b": 1, "c": 3},
    )
    check.equal(
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="unflat"),
        {"a": {"b": 2}, "c": 3},
    )
    with pytest.raises(ValueError, match="duplicated key 'a.b'"):
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="error")
    with pytest.raises(
        ValueError,
        match=(
            "priority argument must be one of 'flat', 'unflat' or 'error' but "
            "found 'UNKNOWN'"
        ),
    ):
        clean_pre_flat({"a.b": 1, "a": {"b": 2}, "c": 3}, priority="UNKNOWN")


def test_save_load_dict() -> None:
    """Test save_dict and load_dict."""
    dict1 = {"a": 1, "b": {"c": 2}, "d": [2, 3.0], "e": [{"f": 4}]}
    save_dict(dict1, "tests/tmp/config.yaml")
    check.is_true(os.path.isfile("tests/tmp/config.yaml"))
    dict2 = load_dict("tests/tmp/config.yaml")
    check.equal(dict1, dict2)
    # Case multiple files and yaml tags
    out_dict = load_dict("tests/configs/multi_files_with_tags.yaml")
    expected_dict = {
        "config@cfg": {
            "param1@par@other": 1,
            "param2": {"a": 1.0, "b@par2": "2.1"},
        },
        "config2": {
            "param3@par3": 3.2,
            "param4@par4": [4, 5, {"6": 6}],
            "param5": [True, 8],
        },
        "config3@cfg3": {"param6": {"param7@par7": None}, "param8": "True"},
        "config4@cfg4": {"config5@cfg5": {"param9": "11"}},
    }
    check.equal(out_dict, expected_dict)
    shutil.rmtree("tests/tmp")


def test_show_dict() -> None:
    """Test show_dict."""
    in_dict = {
        "model": {
            "s1_ae_config": {
                "in_dim": 2,
                "out_dim": 1,
                "layer_channels": [16, 32, 64],
                "conv_per_layer": 1,
                "residual": False,
                "dropout_rate": 0.0,
            },
            "mask_module_dim": [6, 2],
            "glob_module_dims": [2, 8, 2],
            "conv_block_dims": [32, 64, 128],
        },
        "train": {
            "n_epochs": 100,
            "optimizer": {
                "name": "Adam",
                "lr": 0.001,
                "weight_decay": 0.0,
                "warmup_steps": 0,
            },
        },
        "data": {
            "dataset": "mnist",
            "batch_size": 128,
            "num_workers": 6,
        },
    }
    show_dict(in_dict)



"""Functions to create new processing quickly."""
# pylint: disable=unused-argument
import re
from typing import Any, Callable, Dict, Optional, Set

from cliconfig.base import Config
from cliconfig.processing.base import Processing
from cliconfig.tag_routines import clean_all_tags, clean_tag, is_tag_in


class _ProcessingValue(Processing):
    """Processing class for make_processing_value."""

    def __init__(
        self,
        regex: Optional[str],
        tag_name: Optional[str],
        order: float,
        func: Callable[[Any], Any],
    ) -> None:
        super().__init__()
        self.premerge_order = order
        self.regex = regex
        self.tag_name = tag_name
        self.func = func
        if self.regex:
            self.is_in_func = (
                lambda key: re.match(self.regex, key.split(".")[-1]) is not None
            )
        elif self.tag_name:
            self.is_in_func = lambda key: is_tag_in(key, str(self.tag_name))

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if self.is_in_func(flat_key):  # type: ignore
                if self.tag_name:
                    del flat_config.dict[flat_key]
                    flat_key = clean_tag(flat_key, self.tag_name)
                flat_config.dict[flat_key] = self.func(value)
        return flat_config


class _ProcessingValuePersistent(Processing):
    """Processing class for make_processing_value. Persistent version."""

    def __init__(
        self,
        regex: Optional[str],
        tag_name: Optional[str],
        order: float,
        func: Callable[[Any], Any],
    ) -> None:
        super().__init__()
        self.premerge_order = order
        self.regex = regex
        self.tag_name = tag_name
        self.func = func
        self.matched_keys: Set[str] = set()
        if self.regex is not None:
            self.is_in_func = (
                lambda key: re.match(self.regex, key.split(".")[-1]) is not None
            )
        elif self.tag_name is not None:
            self.is_in_func = lambda key: is_tag_in(key, str(self.tag_name))

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if (
                self.is_in_func(flat_key)  # type: ignore
                or clean_all_tags(flat_key) in self.matched_keys
            ):
                self.matched_keys.add(clean_all_tags(flat_key))
                if self.tag_name:
                    del flat_config.dict[flat_key]
                    flat_key = clean_tag(flat_key, self.tag_name)
                flat_config.dict[flat_key] = self.func(value)
        return flat_config


class _ProcessingKeepProperty(Processing):
    """Processing class for make_processing_keep_property."""

    def __init__(
        self,
        regex: Optional[str],
        tag_name: Optional[str],
        premerge_order: float,
        postmerge_order: float,
        func: Callable,
    ) -> None:
        super().__init__()
        self.premerge_order = premerge_order
        self.postmerge_order = postmerge_order
        self.regex = regex
        self.tag_name = tag_name
        self.func = func
        self.properties: Dict[str, Any] = {}
        if self.regex is not None:
            self.is_in_func = (
                lambda key: re.match(self.regex, key.split(".")[-1]) is not None
            )
        elif self.tag_name is not None:
            self.is_in_func = lambda key: is_tag_in(key, str(self.tag_name))

    def premerge(self, flat_config: Config) -> Config:
        """Pre-merge processing."""
        items = list(flat_config.dict.items())
        for flat_key, value in items:
            if self.is_in_func(flat_key):  # type: ignore
                property_ = self.func(value)
                clean_key = clean_all_tags(flat_key)
                if clean_key not in self.properties:
                    self.properties[clean_key] = property_
                if self.tag_name:
                    del flat_config.dict[flat_key]
                    flat_key = clean_tag(flat_key, self.tag_name)
                    flat_config.dict[flat_key] = value
        return flat_config

    def postmerge(self, flat_config: Config) -> Config:
        """Post-merge processing."""
        for flat_key, value in self.properties.items():
            if flat_key in flat_config.dict:
                property_ = self.func(flat_config.dict[flat_key])
                if property_ != value:
                    raise ValueError(
                        f"Property of key {flat_key} has changed from {value} to "
                        f"{property_} while it is protected by a keep-property "
                        "processing (problem found on post-merge)."
                    )
        return flat_config


def create_processing_value(
    func: Callable,
    regex: Optional[str] = None,
    tag_name: Optional[str] = None,
    order: float = 0.0,
    *,
    persistent: bool = True,
) -> Processing:
    r"""Create a processing object that modifies a value in dict using tag or regex.

    The processing is applied on pre-merge. It triggers when the key matches
    the tag or the regex. The function apply ``flat_dict[key] = func(flat_dict[key])``.
    You must only provide one of tag or regex. If tag is provided, the tag will be
    removed from the key during pre-merge.

    Parameters
    ----------
    func : Callable
        The function to apply to the value to make new_value.
    regex : Optional[str]
        The regex to match the key.
    tag_name : Optional[str]
        The tag (without "@") to match the key. The tag is removed from
        the key after triggering.
    order : int, optional
        The pre-merge order. By default 0.0.
    persistent : bool, optional
        If True, the processing will be applied on all keys that have already
        matched the tag before. By nature, regex processing are always persistent.
        By default, True.

    Raises
    ------
    ValueError
        If both tag and regex are provided or if none of them are provided.

    Returns
    -------
    processing : Processing
        The processing object with the pre-merge method.


    Examples
    --------
    With the following config and 2 processings:

    .. code_block: yaml

        # config.yaml
        neg_number1: 1
        neg_number2: 1
        neg_number3@add1: 1

    ::

        proc2 = create_processing_value(lambda x: -x, regex="neg_number.*", order=0.0)
        proc1 = create_processing_value(lambda x: x + 1, tag_name="add1", order=1.0)

    When config.yaml is merged with an other config, it will be considered
    before merging as:

    ::

        {'number1': -1, 'number2': -1, 'number3': 0}
    """
    if tag_name is not None:
        if regex is not None:
            raise ValueError("You must provide a tag or a regex but not both.")
    else:
        if regex is None:
            raise ValueError(
                "You must provide a tag or a regex (to trigger the value update)."
            )
    if persistent:
        return _ProcessingValuePersistent(regex, tag_name, order, func)
    return _ProcessingValue(regex, tag_name, order, func)


def create_processing_keep_property(
    func: Callable,
    regex: Optional[str] = None,
    tag_name: Optional[str] = None,
    premerge_order: float = 0.0,
    postmerge_order: float = 0.0,
) -> Processing:
    """Create a processing object that keep a property from a value using tag or regex.

    The pre-merge processing looks for keys that match the tag or the regex, apply
    the function func on the value and store the result (= the "property"):
    ``property = func(flat_dict[key])``.
    The post-merge processing will check that the property is the same as the one
    stored during pre-merge. If not, it will raise a ValueError.

    Parameters
    ----------
    func : Callable
        The function to apply to the value to define the property to keep.
    regex : Optional[str]
        The regex to match the key.
    tag_name : Optional[str]
        The tag (without "@") to match the key. The values are modified when
        triggering the pattern ".*@<tag_name>.*" and the tag is removed from the key.
    premerge_order : float, optional
        The pre-merge order, by default 0.0
    postmerge_order : float, optional
        The post-merge order, by default 0.0

    Raises
    ------
    ValueError
        If both tag and regex are provided or if none of them are provided.

    Returns
    -------
    Processing
        The processing object with the pre-merge and post-merge methods.

    Note
    ----

        The property to keep can be updated by using the tag a second time.

    Examples
    --------
    A processing that enforce the types of all the parameters to be constant
    (equal to the type of the first value encountered):

    ::

        create_processing_keep_property(type, regex=".*", premerge_order=15.0,
                                        postmerge_order=15.0)

    A processing that protect parameters tagged with @protect from being changed:

    ::

        create_processing_keep_property(lambda x: x, tag_name="protect",
                                        premerge_order=15.0, postmerge_order=15.0)
    """
    if tag_name is not None:
        if regex is not None:
            raise ValueError("You must provide a tag or a regex but not both.")
    else:
        if regex is None:
            raise ValueError(
                "You must provide a tag or a regex (to trigger the value update)."
            )
    processing = _ProcessingKeepProperty(
        regex, tag_name, premerge_order, postmerge_order, func
    )
    return processing



"""Routines to manipulate the tags on the keys of a dict.

Used by the processing functions.
"""
import copy
import re
from typing import Any, Dict, List, Tuple


def clean_tag(flat_key: str, tag_name: str) -> str:
    """Clean a tag from a flat key.

    It removes all occurrences of the tag with the exact name.

    Parameters
    ----------
    flat_key : str
        The flat key to clean.
    tag_name : str
        The name of the tag to remove, with or without the '@' prefix.

    Returns
    -------
    flat_key : str
        The cleaned flat key.

    Note
    ----
        ``tag_name`` is supposed to be the exact name of the tag.

    Examples
    --------
    ::

        >>> clean_tag('abc@tag.def@tag_2.ghi@tag', 'tag')
        abc.def@tag_2.ghi
    """
    if tag_name[0] == "@":
        tag_name = tag_name[1:]
    # Replace "@tag@other_tag" by "@other_tag"
    parts = flat_key.split(f"@{tag_name}@")
    flat_key = "@".join(parts)
    # Replace "@tag." by "."
    parts = flat_key.split(f"@{tag_name}.")
    flat_key = ".".join(parts)
    # Remove "@tag" at the end of the string
    if flat_key.endswith(f"@{tag_name}"):
        flat_key = flat_key[: -len(f"@{tag_name}")]
    return flat_key


def clean_all_tags(flat_key: str) -> str:
    """Clean all tags from a flat key.

    Parameters
    ----------
    flat_key : str
        The flat key to clean.

    Returns
    -------
    flat_key : str
        The cleaned flat key.
    """
    list_keys = flat_key.split(".")
    for i, key in enumerate(list_keys):
        key = re.sub(r"@.*", "", key)
        list_keys[i] = key
    flat_key = ".".join(list_keys)
    return flat_key


def dict_clean_tags(flat_dict: Dict[str, Any]) -> Tuple[Dict[str, Any], List[str]]:
    """Clean a dict from all tags and return the list of keys with tags.

    Parameters
    ----------
    flat_dict : Dict[str, Any]
        The flat dict to clean.

    Returns
    -------
    clean_dict : Dict[str, Any]
        The cleaned flat dict without tags in the keys.
    tagged_keys : List[str]
        The list of keys with tags that have been cleaned.
    """
    items = list(flat_dict.items())
    clean_dict = copy.deepcopy(flat_dict)
    tagged_keys = []
    for key, value in items:
        if "@" in key:
            del clean_dict[key]
            clean_dict[clean_all_tags(key)] = value
            tagged_keys.append(key)
    return clean_dict, tagged_keys


def is_tag_in(flat_key: str, tag_name: str, *, full_key: bool = False) -> bool:
    """Check if a tag is in a flat key.

    The tag name must be the exact name, with or without the "@".
    It supports the case where there are other tags that are prefixes
    or suffixes of the considered tag.

    Parameters
    ----------
    flat_key : str
        The flat key to check.
    tag_name : str
        The name of the tag to check, with or without the '@' prefix.
    full_key : bool, optional
        If True, check for the full key. If False, check for the last part of
        the flat key (after the last dot) that correspond to the parameter name.
        By default, False.

    Returns
    -------
    bool
        True if the tag is in the flat key, False otherwise.
    """
    if tag_name[0] == "@":
        tag_name = tag_name[1:]
    if not full_key:
        flat_key = flat_key.split(".")[-1]
    is_in = (
        flat_key.endswith(f"@{tag_name}")
        or f"@{tag_name}@" in flat_key
        or f"@{tag_name}." in flat_key
    )
    return is_in



"""Test base class of configuration."""

import pytest_check as check

from cliconfig.base import Config
from cliconfig.processing.base import Processing


def test_config(process_add1: Processing) -> None:
    """Test base class of configuration."""
    config_dict = {"a.b": 1, "b": 2, "c.d.e": 3, "c.d.f": [2, 3]}
    config = Config(config_dict, [process_add1])
    check.equal(config.dict, config_dict)
    check.equal(config.process_list, [process_add1])
    check.equal(repr(config), f"Config({config_dict}, ['ProcessAdd1'])")
    config_dict2 = {"c.d.e": 3, "a.b": 1, "b": 2, "c.d.f": [2, 3]}
    config2 = Config(config_dict2, [process_add1])
    check.equal(config, config2)
    config3 = Config(config_dict2, [process_add1, process_add1])
    check.not_equal(config, config3)
    config4 = Config(config_dict2, [])
    check.not_equal(config, config4)



"""Test helpers to create processing functions."""
import re

import pytest
import pytest_check as check

from cliconfig.base import Config
from cliconfig.processing.create import (
    create_processing_keep_property,
    create_processing_value,
)


def test_create_processing_value() -> None:
    """Test create_processing_value."""
    # Persistent
    proc1 = create_processing_value(
        lambda x: x + 1, tag_name="add1", order=1.0, persistent=True
    )
    proc2 = create_processing_value(
        lambda x: -x, regex="neg_number.*", order=0.0, persistent=True
    )
    in_dict = {"neg_number1": 1, "neg_number2": 1, "neg_number3@add1": 1}
    config = Config(in_dict, [proc1, proc2])
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(in_dict, {"neg_number1": -1, "neg_number2": -1, "neg_number3": 0})
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(in_dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 1})
    # Non persistent
    proc2 = create_processing_value(
        lambda x: -x, regex="neg_number.*", order=0.0, persistent=False
    )
    proc1 = create_processing_value(
        lambda x: x + 1, tag_name="add1", order=1.0, persistent=False
    )
    in_dict = {"neg_number1": 1, "neg_number2": 1, "neg_number3@add1": 1}
    config = Config(in_dict, [proc1, proc2])
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(in_dict, {"neg_number1": -1, "neg_number2": -1, "neg_number3": 0})
    config = config.process_list[1].premerge(config)
    config = config.process_list[0].premerge(config)
    check.equal(in_dict, {"neg_number1": 1, "neg_number2": 1, "neg_number3": 0})

    # Failing cases
    with pytest.raises(
        ValueError, match="You must provide a tag or a regex but not both."
    ):
        create_processing_value(lambda x: x + 1, tag_name="add1", regex="neg_number.*")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "You must provide a tag or a regex (to trigger the value update)."
        ),
    ):
        create_processing_value(lambda x: x + 1, order=0.0)


def test_create_processing_keep_property() -> None:
    """Test create_processing_keep_property."""
    proc1 = create_processing_keep_property(type, regex=".*")
    proc2 = create_processing_keep_property(lambda x: x, tag_name="protect")
    in_dict = {"str": "foo", "int": 2, "list@protect": [1, 2, 3]}
    config = Config(in_dict, [proc1, proc2])
    config = config.process_list[0].premerge(config)
    config = config.process_list[1].premerge(config)
    check.equal(config.dict, {"str": "foo", "int": 2, "list": [1, 2, 3]})
    config = config.process_list[0].postmerge(config)
    config = config.process_list[1].postmerge(config)
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Property of key str has changed from <class 'str'> to <class 'int'> "
            "while it is protected by a keep-property processing (problem found on "
            "post-merge)."
        ),
    ):
        config.process_list[0].postmerge(Config({"str": 0}, []))
    with pytest.raises(
        ValueError,
        match=re.escape(
            "Property of key list has changed from [1, 2, 3] to [1, 2] while "
            "it is protected by a keep-property processing (problem found on "
            "post-merge)."
        ),
    ):
        config.process_list[1].postmerge(Config({"list": [1, 2]}, []))
    config.process_list[0].premerge(Config({"str": 0}, []))  # should not raise
    config.process_list[1].premerge(Config({"list": [1, 2]}, []))  # should not raise

    # Failing cases
    with pytest.raises(
        ValueError, match="You must provide a tag or a regex but not both."
    ):
        create_processing_keep_property(lambda x: x, tag_name="protect", regex=".*")
    with pytest.raises(
        ValueError,
        match=re.escape(
            "You must provide a tag or a regex (to trigger the value update)."
        ),
    ):
        create_processing_keep_property(lambda x: x, premerge_order=0.0)
